/* SHARED SKETCH ALGORITHM FOR LEARNING ACTIVE BASIS -- piecewise version */
# include <stdio.h>
# include <stdlib.h>
# include "mex.h"        
# include "math.h"
# define PI 3.1415926
# define ABS(x) ((x)>0? (x):(-(x)))
# define MAX(x, y) ((x)>(y)? (x):(y))
# define MIN(x, y) ((x)<(y)? (x):(y))
# define ROUND(x) (floor((x)+.5))
# define NEGMAX -1e10
/* Generating float vector */
float *float_vector(int n)
{
    float *v; 
    v = (float*) mxCalloc (n, sizeof(float));    
    return v; 
}
/* Generating integer vector */
int *int_vector(int n)
{
    int *v; 
    v = (int*) mxCalloc (n, sizeof(int));
    return v; 
}
/* Compute pixel index in the vector that stores image */
int px(int x, int y, int lengthx, int lengthy)  /* the image is lengthx*lengthy */
{            
   return (x + (y-1)*lengthx - 1); 
 }
/* key input and output variables */
int locationShiftLimit; /* key parameters */
int numImage, numElement; /* number of images and number of Gabors */   
double saturation; /* saturation level for sigmoid transformation */
float **SUM1map, **MAX1map, *pooledMax1map; /* pooledMax1map is summed over all images */ 
float **trackMap; /* track the shift or perturbation at each location and orientation */
double *Correlation, *allSymbol; /* correlation between Gabors, and symbols of Gabors */              
int halfFilterSize; /* filter size = 2*halfFilterSize + 1 */ 
int sizex, sizey, subsample, sizexSubsample, sizeySubsample; /* MAX1 maps are smaller than SUM1 maps by subsample */ 
int numShift, *xShift, *yShift; /* stored shifts in x and y, and the shifted orientations */      
int numStoredPoint; /* number of stored points of lambda in exponential model */                     
double *storedlambda, *storedExpectation, *storedLogZ; /* stored lambda, mu, logZ in exponential model */    
int startx, endx, starty, endy; /* search within the interior of the image */
double *selectedx, *selectedy, *selectedlambda, *selectedLogZ, *SUM2score; /* parameters of learned active basis */  
float *commonTemplate, **deformedTemplate; /* templates of input images and template matching scores */    
/* store all the shifts or perturbations in local maximum pooling */
/* store all the shifts or perturbations in local maximum pooling */
void StoreShift()  
{
    int shift, r, x, y;
    /* store all the possible shifts for all orientations */
    numShift = (locationShiftLimit*2+1)*(locationShiftLimit*2+1); 
    xShift = int_vector(numShift); 
    yShift = int_vector(numShift); 
    shift = 0; 
    /* order the shifts from small to large */
    for (r=0; r<=2*locationShiftLimit*locationShiftLimit; r++)        
    {
        for (x=-locationShiftLimit; x<=locationShiftLimit; x++)
            for (y=-locationShiftLimit; y<=locationShiftLimit; y++)
              {
                if (ABS(x*x+y*y-r)<.1)
                {
                    xShift[shift] = x*subsample; 
                    yShift[shift] = y*subsample; 
                    shift ++;
                }
             }
    }
 }
/* local maximum pooling at (x, y, orient) */
float LocalMaximumPooling(int img, int x, int y, int *trace) 
{
   float maxResponse, r;
   int shift, x1, y1, mshift; 
 
   maxResponse = NEGMAX; mshift = 0;  
   for (shift=0; shift<numShift; shift++)
   {
       x1 = x + xShift[shift]; 
       y1 = y + yShift[shift]; 
       if ((x1>=1)&&(x1<=sizex)&&(y1>=1)&&(y1<=sizey))
        {            
           r = SUM1map[img][px(x1, y1, sizex, sizey)]; 
           if (r>maxResponse)
                {
                   maxResponse = r;   
                   mshift = shift; 
                }
         }
     }
   trace[0] = mshift;  /* return the shift in local maximum pooling */
   return(maxResponse); 
}
/* the local maximal Gabor inhibits overlapping Gabors */
void NonMaximumSuppression(int img, int mx, int my) 
{
   int x, y,  x1, y1, here, shift, startx0, endx0, starty0, endy0, trace[2]; 
   float maxResponse, maxResponse1; 
   /* inhibit on the SUM1 maps */  
    for (x=MAX(1, mx-2*halfFilterSize); x<=MIN(sizex, mx+2*halfFilterSize); x++)
         for (y=MAX(1, my-2*halfFilterSize); y<=MIN(sizey, my+2*halfFilterSize); y++)
         {
          SUM1map[img][px(x, y, sizex, sizey)] *= 
          Correlation[px(x-mx+2*halfFilterSize+1, y-my+2*halfFilterSize+1, 4*halfFilterSize+1, 4*halfFilterSize+1)];
         }
   /* update the MAX1 maps */
   startx0 = floor((mx-2*halfFilterSize)/subsample)-locationShiftLimit+1; 
   starty0 = floor((my-2*halfFilterSize)/subsample)-locationShiftLimit+1; 
   endx0 =   floor((mx+2*halfFilterSize)/subsample)+locationShiftLimit; 
   endy0 =   floor((my+2*halfFilterSize)/subsample)+locationShiftLimit; 
   for (x=MAX(startx, startx0); x<=MIN(endx, endx0); x++)
         for (y=MAX(starty, starty0); y<=MIN(endy, endy0); y++)
         { /* go over the locations that may be affected */         
           here = px(x, y, sizexSubsample, sizeySubsample);        
           maxResponse = MAX1map[img][here]; 
           shift = ROUND(trackMap[img][here]);
           x1 = x*subsample + xShift[shift]; 
           y1 = y*subsample + yShift[shift];            
           if ((x1-mx>=-2*halfFilterSize)&&(x1-mx<=2*halfFilterSize)&&
               (y1-my>=-2*halfFilterSize)&&(y1-my<=2*halfFilterSize)) 
             { /* if the previous local maximum is within the inhibition range */
              if(Correlation[px(x1-mx+2*halfFilterSize+1,y1-my+2*halfFilterSize+1,4*halfFilterSize+1,4*halfFilterSize+1)]==0.) 
                 {   /* if it is indeed inhibited */
                     maxResponse1 = LocalMaximumPooling(img, x*subsample, y*subsample, trace); 
                     trackMap[img][here] = trace[0];  
                     MAX1map[img][here] = maxResponse1;
                     pooledMax1map[here] += (maxResponse1-maxResponse);                  
                 }
             }         
         }               
}
/* plot the bar for Gabor at (mx, my, mo) */
void DrawElement(float *Template, int mx, int my, double w) 
{
  int x, y, here; 
  float a; 
          
  for (x=mx-halfFilterSize; x<=mx+halfFilterSize; x++)
     for (y=my-halfFilterSize; y<=my+halfFilterSize; y++)
      if ((x>=1)&&(x<=sizex)&&(y>=1)&&(y<=sizey)) 
        {
         a = allSymbol[px(x-mx+halfFilterSize+1, y-my+halfFilterSize+1, 
                              2*halfFilterSize+1, 2*halfFilterSize+1)]*w; 
         here = px(x, y, sizex, sizey); 
         if (Template[here]<a)
             Template[here] = a;
        }
}
/* Initialize MAX1 maps */
void InitializepooledMAX1map()
{
   int img, x, y, here;  
   /* calculate MAX1 maps and record the shifts */
   pooledMax1map = float_vector(sizexSubsample*sizeySubsample);  
   startx = floor((halfFilterSize+1)/subsample)+1+locationShiftLimit; 
   endx = floor((sizex-halfFilterSize)/subsample)-1-locationShiftLimit; 
   starty = floor((halfFilterSize+1)/subsample)+1+locationShiftLimit; 
   endy = floor((sizey-halfFilterSize)/subsample)-1-locationShiftLimit; 
   for (x=startx; x<=endx; x++)
      for (y=starty; y<=endy; y++)
       {
        here = px(x, y, sizexSubsample, sizeySubsample);     
        pooledMax1map[here] = 0.; 
        for (img=0; img<numImage; img++)
             pooledMax1map[here] += MAX1map[img][here];     
       }
}
/* Shared sketch algorithm for pursuing Gabors */
void PursueElement()
{
   int img, x, y, bestx, besty, mx, my, t, here, shift, j; 
   float r, maxPooled, maxResponse, average, overShoot; 

   t = 0; /* t is for iterations */
   do  
   { /* select the next Gabor */
     maxPooled = NEGMAX;  
     for (x=startx; x<=endx; x++)
      for (y=starty; y<=endy; y++)
       {
        here = px(x, y, sizexSubsample, sizeySubsample); 
        r = pooledMax1map[here]; 
        if (maxPooled<r)
         {
           maxPooled = r; 
           bestx = x; besty = y; 
         }
       }
     /* estimate the parameter of exponential model */
     selectedx[t] = bestx; selectedy[t] = besty; 
     average = maxPooled/numImage;    
     j = numStoredPoint-1; 
     while (storedExpectation[j]>average)
         j--; 
     if (j==numStoredPoint-1)
      {
        selectedlambda[t] = storedlambda[j]; 
        selectedLogZ[t] = storedLogZ[j]; 
       }
     else 
      { /* linear interpolation */
        overShoot = (average-storedExpectation[j])/(storedExpectation[j+1]-storedExpectation[j]); 
        selectedlambda[t] = storedlambda[j]+(storedlambda[j+1]-storedlambda[j])*overShoot; 
        selectedLogZ[t] = storedLogZ[j]+(storedLogZ[j+1]-storedLogZ[j])*overShoot; 
      }  
     /* plot selected and perturbed Gabor and inhibit nearby Gabors */
     DrawElement(commonTemplate, bestx*subsample, besty*subsample, sqrt(average)); 
     here = px(bestx, besty, sizexSubsample, sizeySubsample); 
     for (img=0; img<numImage; img++)
        {    
          maxResponse = MAX1map[img][here]; 
          SUM2score[img] += (selectedlambda[t]*maxResponse - selectedLogZ[t]); 
          shift = ROUND(trackMap[img][here]); 
          mx = bestx*subsample + xShift[shift]; 
          my = besty*subsample + yShift[shift];      
         if (maxResponse>0.)
            {  
               DrawElement(deformedTemplate[img], mx, my, sqrt(maxResponse)); 
               NonMaximumSuppression(img, mx, my);  
             } 
         } 
      printf("%d:  (x = %d   y = %d) --> %f\n", t, bestx, besty, selectedlambda[t]); 
      t++; 
   }
  while (t<numElement);  
}
/* read in input variables and run the algorithm */
void mexFunction(int nlhs, mxArray *plhs[], 
                 int nrhs, const mxArray *prhs[])                
{
 int img, c; 
 mxArray *f;  
 
 c = 0; 
 /* about active basis */
 locationShiftLimit = ROUND(mxGetScalar(prhs[c++]));    
 saturation = mxGetScalar(prhs[c++]); 
 subsample = ROUND(mxGetScalar(prhs[c++]));   
 numElement = ROUND(mxGetScalar(prhs[c++])); 
 /* about input images */
 numImage = ROUND(mxGetScalar(prhs[c++])); 
 sizex = ROUND(mxGetScalar(prhs[c++])); 
 sizey = ROUND(mxGetScalar(prhs[c++]));   
 SUM1map = mxCalloc(numImage, sizeof(float*));  
 for (img=0; img<numImage; img++)
      {  
       f = mxGetCell(prhs[c], img); 
       SUM1map[img] = mxGetPr(f);    
      }
 c++; 
 MAX1map = mxCalloc(numImage, sizeof(float*));  
 for (img=0; img<numImage; img++)
      {  
       f = mxGetCell(prhs[c], img); 
       MAX1map[img] = mxGetPr(f);    
      }
 c++; 
 trackMap = mxCalloc(numImage, sizeof(float*));  
 for (img=0; img<numImage; img++)
      {  
       f = mxGetCell(prhs[c], img); 
       trackMap[img] = mxGetPr(f);    
      }
 c++; 
 /* about Gabor filters */
 halfFilterSize = ROUND(mxGetScalar(prhs[c++]));   
 Correlation = mxGetPr(prhs[c++]);   
 allSymbol = mxGetPr(prhs[c++]);    
 /* about exponential model */ 
 numStoredPoint = ROUND(mxGetScalar(prhs[c++])); 
 storedlambda = mxGetPr(prhs[c++]);   
 storedExpectation = mxGetPr(prhs[c++]);   
 storedLogZ = mxGetPr(prhs[c++]);  
 /* learned parameters of active basis */         
 selectedx = mxGetPr(prhs[c++]);         
 selectedy = mxGetPr(prhs[c++]);   
 selectedlambda = mxGetPr(prhs[c++]);       
 selectedLogZ = mxGetPr(prhs[c++]);  
 SUM2score = mxGetPr(prhs[c++]);  
 /* templates of images */
 commonTemplate = mxGetPr(prhs[c++]);     
 deformedTemplate = mxCalloc(numImage, sizeof(float*));   
 for (img=0; img<numImage; img++)
    {
        f = mxGetCell(prhs[c], img);
        deformedTemplate[img] = mxGetPr(f);  
     }
 c++; 
 /* MAX1 maps are smaller than SUM1 maps */
 sizexSubsample = floor((double)sizex/subsample); 
 sizeySubsample = floor((double)sizey/subsample); 
 /* run the shared sketch algorithm */
 StoreShift();  
 InitializepooledMAX1map(); 
 PursueElement(); 
 /* free matrices to avoid memory leak */  
 mxFree(pooledMax1map);  
 mxFree(xShift); 
 mxFree(yShift); 
}



